/***************************************************************************
  
  e-puck_drive -- E-puck robot evolves to drive in an arena and avoid walls
  This program is free software; any publications presenting results
  obtained with this program must mention it and its origin. You 
  can redistribute it and/or modify it under the terms of the GNU 
  General Public License as published by the Free Software 
  Foundation; either version 2 of the License, or (at your option) 
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

***************************************************************************/

#include "e-puck_drive.h"

int main()
{
  robot_live(reset);
  robot_run(run);
  return 0;
}

//
// Reset the robot controller and initiate the sensors and emitter/receiver
//
static void reset(void)
{  
  int i;
  mode =1;
  emitter = robot_get_device("emitterepuck");
  buffer = (float *) emitter_get_buffer(emitter);
 
  receiver = robot_get_device("receiver");
  
  receiver_enable(receiver, TIME_STEP);
  
  // Create a supervised node for the robot 
  //const char *robot_name = robot_get_name();
  //robot = supervisor_node_get_from_def(robot_name);
  
  char text[5]="led0";
  for(i=0;i<NB_LEDS;i++) {
    led[i]=robot_get_device(text); // get a handler to the sensor 
    text[3]++; // increase the device name to "ps1", "ps2", etc. 
  }
  
  text[0]='p';
  text[1]='s';
  text[3]='\0';
  
  text[2]='0';
  ps[0] = robot_get_device(text); // proximity sensors
  text[2]='7';
  ps[1] = robot_get_device(text); // proximity sensors
  text[2]='1';
  ps[2] = robot_get_device(text); // proximity sensors
  text[2]='6';
  ps[3] = robot_get_device(text); // proximity sensors
  text[2]='2';
  ps[4] = robot_get_device(text); // proximity sensors
  text[2]='5';
  ps[5] = robot_get_device(text); // proximity sensors
  text[2]='3';
  ps[6] = robot_get_device(text); // proximity sensors
  text[2]='4';
  ps[7] = robot_get_device(text); // proximity sensors

  // Enable proximity and floor sensors
  for(i=0;i<NB_DIST_SENS;i++) {
    distance_sensor_enable(ps[i],TIME_STEP);
  }
  
  // Enable GPS sensor to determine position
  gps=robot_get_device("gps");
  gps_enable(gps, TIME_STEP);
  
  return;
}

//------------------------------------------------------------------------------
//
//    CONTROLLER
//
//------------------------------------------------------------------------------

////////////////////////////////////////////
// Main
static int run(int ms)
{
  int i;
  if (mode!=robot_get_mode())
  {
    mode = robot_get_mode();
    if (mode == SIMULATION) {
      for(i=0;i<NB_DIST_SENS;i++) ps_offset[i]=PS_OFFSET_SIMULATION[i];
      robot_console_printf("Switching to SIMULATION.\n\n");
    } 
    else if (mode == REALITY) { 
      for(i=0;i<NB_DIST_SENS;i++) ps_offset[i]=PS_OFFSET_REALITY[i];
      robot_console_printf("\nSwitching to REALITY.\n\n");
    }
  }

  // if we're testing a new genome, receive weights and initialize trial
  if (step == 0) {
    int n = receiver_get_queue_length(receiver);
    //wait for new genome
    if (n) {
      const float *genes = (float *) receiver_get_data(receiver);
      //set neural network weights
      for (i=0;i<NB_WEIGHTS;i++) weights[i]=genes[i];
      receiver_next_packet(receiver);
    }
    else return TIME_STEP;
    
    const float *gps_matrix = gps_get_matrix(gps);
    robot_initial_position[0] = gps_position_x(gps_matrix);
    robot_initial_position[1] = gps_position_z(gps_matrix);
    fitness = 0;
  }
  
  step++;
  //robot_console_printf("Step: %d\n", step);
    
  if(step < TRIAL_DURATION/(float)TIME_STEP) {
    //drive robot
    fitness+=run_trial();
    //send message with current fitness
    float msg[2] = {fitness, 0.0};
    emitter_send_packet(emitter, (void *)msg, 2*sizeof(float));
  }
  else {
    //send message to indicate end of trial
    float msg[2] = {fitness, 1.0};
    emitter_send_packet(emitter, (void *)msg, 2*sizeof(float));
    //reinitialize counter
    step = 0;
  }
  
  return TIME_STEP;
  return TIME_STEP;
}

//
// Trial to test the robot's behavior (according to NN) in the environment
//
float run_trial() {

  float inputs[NB_INPUTS];
  float outputs[NB_OUTPUTS];
  
  //Get position of the e-puck
  static float position[3]={0.0,0.0,0.0};
  const float *gps_matrix = gps_get_matrix(gps);
  position[0] = gps_position_x(gps_matrix);
  position[1] = gps_position_z(gps_matrix);
  position[2] = gps_position_y(gps_matrix);
  
  //Speed of the robot's wheels within the +SPEED_RANGE and -SPEED_RANGE values
  int speed[2]={0,0};
  //Maximum activation of all IR proximity sensors [0,1]
  float maxIRActivation = 0;

  //get sensor data, i.e., NN inputs
  int i, j;
  for(j=0;j<NB_INPUTS;j++) {
    inputs[j]=(((float)distance_sensor_get_value(ps[j])-ps_offset[j])<0)?0:((float)distance_sensor_get_value(ps[j])-ps_offset[j])/((float)PS_RANGE);
    //get max IR activation
    if(inputs[j]>maxIRActivation) maxIRActivation=inputs[j];
  }
  //robot_console_printf("max IR activation: %f\n", maxIRActivation);

  // Run the neural network and computes the output speed of the robot
  run_neural_network(inputs, outputs); 
  
  speed[LEFT]  = SPEED_RANGE*outputs[0];
  speed[RIGHT] = SPEED_RANGE*outputs[1];
  
  // If you are running against an obstacle, stop the wheels
  if ( maxIRActivation > 0.9 ) {
    speed[LEFT]=0;
    speed[RIGHT]=0;
  }
  // Set wheel speeds to output values
  differential_wheels_set_speed(speed[LEFT], speed[RIGHT]); //left, right
  
  // Stop the robot if it is against an obstacle
  for (i=0;i<NB_DIST_SENS;i++) {
    int tmpps=(((float)distance_sensor_get_value(ps[i])-ps_offset[i])<0)?0:((float)distance_sensor_get_value(ps[i])-ps_offset[i]);
    
    if(OBSTACLE_THRESHOLD<tmpps ) {// proximity sensors 
      //robot_console_printf("%d \n",tmpps);
      speed[LEFT]  = 0;
      speed[RIGHT] = 0;
      break;
    }
  }
  
  return compute_fitness(speed, position, maxIRActivation);
}

//
// Computes the fitness of the robot at each timestep.
//
float compute_fitness(int speed[2], float position[3], float maxIRActivation)
{
  /*
  Input parameters:
  
  speed[0] //leftSpeed
  speed[1] //rightSpeed
  
  SPEED_RANGE //all speed outputs are within the +SPEED_RANGE and -SPEED_RANGE values
  
  position[0] //current X coordinate
  position[1] //current Y coordinate
  
  robot_initial_position[0] //initial X coordinate
  robot_initial_position[1] //initial Y coordinate
  
  maxIRActivation //highest IR sensor activation [0,1]
  */
  
  return 0.0;
}

//
// Run the neural network
//
void run_neural_network(float* inputs, float* outputs)
{
  int i,j;
  int weight_counter=0;
  
  if(NB_HIDDEN_NEURONS>0){
    float hidden_neuron_out[(int)NB_HIDDEN_NEURONS];
    
    for(i=0;i<NB_HIDDEN_NEURONS;i++) {
      float sum=0;
      for(j=0;j<NB_INPUTS;j++) {
        sum+=inputs[j]*weights[weight_counter];
        weight_counter++;
      }
      hidden_neuron_out[i]=tanh(sum+weights[weight_counter]);
      weight_counter++;
    }
    
    for(i=0;i<NB_OUTPUTS;i++) {
      float sum=0;
      for(j=0;j<NB_HIDDEN_NEURONS;j++)
      {
        sum+=hidden_neuron_out[j]*weights[weight_counter];
        weight_counter++;
      }
      outputs[i]=tanh(sum+weights[weight_counter]);
      weight_counter++;
    }
  }
  else
  {
    for(i=0;i<NB_OUTPUTS;i++) {
      float sum=0.0;
      for(j=0;j<NB_INPUTS;j++)
      {
        sum+=inputs[j]*weights[weight_counter];
        weight_counter++;
      } 
      outputs[i]=tanh(sum+weights[weight_counter]);
      weight_counter++;
    }
  }

  return;
}

