/*
 * File:         evolution.c
 */

#include "evolution.h"

int main()
{
  robot_live(reset);
  robot_run(run);             /* never returns */
  return 0;
}

//
//The reset function is called at the beginning of an evolution. 
//
static void reset(void)
{
  int i;
  for(i=0; i<POP_SIZE; i++) fitness[i]=-1;
  srand(time(0));  
  
  // Initiate the emitter used to send genomes to the experiment
  emitter = robot_get_device("emitter");
  buffer = (float *) emitter_get_buffer(emitter);
 
  // Initiate the receiver used to receive fitness
  receiver = robot_get_device("receiversupervisor");
  receiver_enable(receiver, TIME_STEP);
  
  // Create a supervised node for the robot 
  //const char *robot_name = robot_get_name();
  //robot = supervisor_node_get_from_def(robot_name);
  robot = supervisor_node_get_from_def("EPUCK");

  /*
  * The call to this function is mandatory for the function
  * supervisor_node_was_found to be able to work correctly.
  */
  robot_step(0);

  // set the robot controller to nn
  const char *controller_name = "e-puck_drive";
  supervisor_robot_set_controller(robot, controller_name);
  
  //check whether robot was found
  if (supervisor_node_was_found(robot) != 1) 
    robot_console_printf("Error: node EPUCK not found!!!\n");

  if(EVOLVING) {
    //Open log files
    f1= fopen ("../../data/fitness.txt", "wt");
    f2= fopen ("../../data/genomes.txt", "wt");
    //initial weights randomly
    initializePopulation();
    robot_console_printf("NEW EVOLUTION\n");
    robot_console_printf("GENERATION 0\n");
    // send genomes to experiment
    resetRobotPosition();
    emitter_send_packet(emitter, (void *)pop[evaluated_inds], NB_GENES*sizeof(float));
  }
  else { // testing best individual
    //Read best genome from bestgenome.txt and initialize weights.
    f3= fopen ("../../data/bestgenome.txt", "rt");
    fscanf(f3,"%d %d", &generation, &evaluated_inds);
    for(i=0;i<NB_GENES;i++) fscanf(f3,"%f ",&pop[0][i]);
    
    robot_console_printf("TESTING INDIVIDUAL %d, GENERATION %d\n", evaluated_inds, generation);

    // send genomes to experiment
    resetRobotPosition();
    emitter_send_packet(emitter, (void *)pop[0], NB_GENES*sizeof(float));
  }
  return;
}

//
//main controller of the evolution, this function never returns
//
static int run(int ms)
{
  const float *fit;
  float finished=0;

  // as long as individual is being evaluated, print current fitness and return
  int n = receiver_get_queue_length(receiver);
  if (n) {
    fit = (float *)receiver_get_data(receiver);
    fitness[evaluated_inds]=fit[0];
    finished=fit[1];
    // print stuff on screen
    char* message;
    sprintf(message, "Gen: %d Ind: %d Fit: %.2f", generation, evaluated_inds, fit[0]);
    supervisor_set_label(0, message, 0, 0, 0.1, 0x000000);
    receiver_next_packet(receiver);
  }

  // when evaluation is done, an extra flag is returned in the message
  if (!finished) return TIME_STEP;

  if(EVOLVING) {
    // if whole population has been evaluated
    if ((evaluated_inds+1) == POP_SIZE ) {  
      // sort population by fitness
      sortPopulation();
      // find and log current and absolute best individual
      bestfit=sortedfitness[0][0];
      bestind=(int)sortedfitness[0][1];
      if (bestfit > abs_bestfit) {
        abs_bestfit=bestfit;
        abs_bestind=bestind;
        logBest();
      }
      robot_console_printf("best fit: %f\n", bestfit);
      
      //write data to files
      logPopulation();
      
      //rank population, select best individuals and create new generation
      createNewPopulation();
      
      generation++;
      robot_console_printf("\nGENERATION %d\n", generation);
      evaluated_inds = 0;
      avgfit = 0.0;
      bestfit = -1.0;
      bestind = -1.0;
      
      resetRobotPosition();
      emitter_send_packet(emitter, (void *)pop[evaluated_inds], NB_GENES*sizeof(float));
    }
    else {
      // assign received fitness to individual
      robot_console_printf("fitness: %f\n", fitness[evaluated_inds]);
      evaluated_inds++;
      
      // send next genome to experiment
      resetRobotPosition();
      emitter_send_packet(emitter, (void *)pop[evaluated_inds], NB_GENES*sizeof(float));
    }
  }
  
  return TIME_STEP;
}

//
//Initiaite genes of all individuals randomly
//
void initializePopulation(void)
{
  int i, j;
  for(i=0;i<POP_SIZE;i++) {
    for(j=0;j<NB_GENES;j++) {
      // all genes must be in the range of [-1, 1]
      pop[i][j]=(GENE_MAX-GENE_MIN)*(float)rand()/(float)RAND_MAX
               -(GENE_MAX-GENE_MIN)/2.0;
    }
  }
  return;
}

//
//resets the position of the robot before each trial
//
void resetRobotPosition(void)
{
  if (supervisor_node_was_found(robot) != 1)
    robot_console_printf("Error: node %s not found\n", "EPUCK");

  //Initial position of the robot at each time step..
  //float* robot_initial_position =  float[3];
  int i;
  for (i=0; i<2; i++) robot_initial_position[i] = 0.6*(float)rand()/(float)RAND_MAX-0.3;
  robot_initial_position[2] = 6.28*(float)rand()/(float)RAND_MAX;
  //robot_initial_position[3] = 10.0;
  supervisor_field_set(robot, SUPERVISOR_FIELD_TRANSLATION_X | SUPERVISOR_FIELD_TRANSLATION_Z | SUPERVISOR_FIELD_ROTATION_ANGLE,robot_initial_position);
  //supervisor_simulation_physics_reset();

  return;
}

//
//get the position of the robot at any time step
//
/*
float* getRobotPosition(void)
{
  float position[3];
  supervisor_field_get(robot,SUPERVISOR_FIELD_TRANSLATION_X | SUPERVISOR_FIELD_TRANSLATION_Z | SUPERVISOR_FIELD_ROTATION_ANGLE, &position, TIME_STEP);
  return position;
}
*/

//
//Based on the fitness of the last generation, generate a new population of 
//genomes for the next generation. 
//
void createNewPopulation(void)
{
  float newpop[POP_SIZE][NB_GENES];
  int elitism_counter = POP_SIZE*ELITISM_RATIO;
  float total_fitness = 0;
  // find minimum fitness to subtract it from sum
  float min_fitness = sortedfitness[POP_SIZE-1][0];
  if (min_fitness<0) min_fitness=0;
  int i, j;
  
  // calculate total of fitness, used for roulette wheel selection
  for(i=0; i<POP_SIZE; i++) total_fitness+=fitness[i];
  total_fitness-=min_fitness*POP_SIZE;
  
  //create new population
  for(i=0; i<POP_SIZE; i++) {
  
    //the elitism_counter best individuals are simply copied to the new population
    if(i<elitism_counter) {
      for(j=0;j<NB_GENES;j++) 
        newpop[i][j]=pop[(int)sortedfitness[i][1]][j];
    }
    //the other individuals are generated through the crossover of two parents
    else {
    
      //select non-elitist individual
      int ind1=0;
      if (ROULETTE_WHEEL==1) {
        float r=(float)rand()/(float)RAND_MAX;
        float fitness_counter=(sortedfitness[ind1][0]-min_fitness)/total_fitness;
        while( r > fitness_counter) { 
          ind1++; 
          fitness_counter+=(sortedfitness[ind1][0]-min_fitness)/total_fitness; 
        }
      }
      else ind1=elitism_counter+((float)rand()/(float)RAND_MAX)*(POP_SIZE*REPRODUCTION_RATIO-elitism_counter);
      
      //if we will do crossover, select a second individual
      if ((float)rand()/(float)RAND_MAX < CROSSOVER_PROBABILITY) { 
        int ind2=0;
        if (ROULETTE_WHEEL==1)
          do {
            float r=(float)rand()/(float)RAND_MAX;
            float fitness_counter=(sortedfitness[ind2][0]-min_fitness)/total_fitness;
            while( r > fitness_counter) { 
              ind2++; 
              fitness_counter+=(sortedfitness[ind2][0]-min_fitness)/total_fitness; 
            }
          } while (ind1==ind2);
        else 
          do {
            ind2=elitism_counter+((float)rand()/(float)RAND_MAX)*(POP_SIZE*REPRODUCTION_RATIO-elitism_counter);
          } while (ind1==ind2);
        ind1=(int)sortedfitness[ind1][1];
        ind2=(int)sortedfitness[ind2][1];
        crossover(ind1, ind2, newpop[i]);
      }
      else { //if no crossover was done, just copy selected individual directly
        for(j=0;j<NB_GENES;j++) newpop[i][j]=pop[(int)sortedfitness[ind1][1]][j];
      }
    }
  }
  
  //mutate new population and copy back to pop
  for(i=0; i<POP_SIZE; i++) {
    if(i<elitism_counter) { //no mutation for elitists
      for(j=0;j<NB_GENES;j++) 
        pop[i][j]=newpop[i][j];
    }
    else { //mutate others with probability per gene
      for(j=0;j<NB_GENES;j++) 
        if((float)rand()/(float)RAND_MAX<MUTATION_PROBABILITY)
          pop[i][j]=mutate(GENE_MIN,GENE_MAX,newpop[i][j]); 
        else
          pop[i][j]=newpop[i][j];
    }
    
    //reset fitness
    fitness[i]=-1;
  }
  return;
}

//
//Mutate a given gene value within the min-max bounds
//
float mutate(float min, float max, float gene)
{
  float  x1, x2, w, y1;
 
  do {
    x1 = GENE_MAX-GENE_MIN * ((float)rand()/(float)RAND_MAX) - (GENE_MAX-GENE_MIN)/2.0;
    x2 = GENE_MAX-GENE_MIN * ((float)rand()/(float)RAND_MAX) - (GENE_MAX-GENE_MIN)/2.0;
    w = x1 * x1 + x2 * x2;
  } while ( w > 1.0 || w == 0 );
 
  y1 = gene + MUTATION_SIGMA * x1 * sqrt( (-2.0 * log( w ) ) / w );

  if (y1>max) return max;
  if (y1<min) return min;
  return y1;
}

//
//Crossover operator for two individuals ind1 and ind2. Returns crossed individual
//
void crossover(int ind1, int ind2, float* new_ind) {
  int crossover_point = 0, i;
  float nb_genes=NB_GENES;

  crossover_point=floor(nb_genes*(float)rand()/(float)RAND_MAX);
  
  for(i=0;i<NB_GENES;i++) { 
    if(i<=crossover_point) new_ind[i]=pop[ind1][i];
    else                   new_ind[i]=pop[ind2][i];
  }
}

//
//Sort population by fitness
//
void sortPopulation() {
  int i;
  //sort population by fitness
  for (i=0; i<POP_SIZE; i++) {
    sortedfitness[i][0]=fitness[i];
    sortedfitness[i][1]=(float)i; //keep index
  }
  quickSort(sortedfitness, 0, POP_SIZE - 1);
}

//
// Standard fast algorithm to sort population by fitness
//
void quickSort(float fitness[][2], int left, int right)
{
  float pivot[2];
  int l_hold, r_hold;
 
  l_hold = left;
  r_hold = right;
  pivot[0] = fitness[left][0];
  pivot[1] = fitness[left][1];
  while (left < right)
  {
    while ((fitness[right][0] <= pivot[0]) && (left < right))
      right--;
    if (left != right) {
      fitness[left][0] = fitness[right][0];
      fitness[left][1] = fitness[right][1];
      left++;
    }
    while ((fitness[left][0] >= pivot[0]) && (left < right))
      left++;
    if (left != right) {
      fitness[right][0] = fitness[left][0];
      fitness[right][1] = fitness[left][1];
      right--;
    }
  }
  fitness[left][0] = pivot[0];
  fitness[left][1] = pivot[1];
  pivot[0] = left;
  left = l_hold;
  right = r_hold;
  if (left < (int)pivot[0])  quickSort(fitness, left, (int)pivot[0]-1);
  if (right > (int)pivot[0]) quickSort(fitness, (int)pivot[0]+1, right);
}

//
//Write all genomes and fitnesses to file
//
void logPopulation() {
  int i, j;
  //calculate average fitness
  for (i=0; i<POP_SIZE; i++) {
    avgfit+=fitness[i];
  }
  avgfit/=(float)POP_SIZE;
  
  //fitness
  fprintf(f1, "%d\t %f\t %f\n", 
              generation, avgfit, bestfit);
  fflush(f1);
  
  //all genomes
  fprintf(f2, "gen: %d\n", generation);
  for(i=0; i<POP_SIZE; i++) {
    fprintf(f2, "ind: %d\t fitness: %f\tgenes:\t", i, fitness[i]);
    for(j=0; j<NB_GENES; j++)
      fprintf(f2, "%f\t", pop[i][j]);
    fprintf(f2, "\n");
    fflush(f2);
  }
  fprintf(f2, "best ind: %d\tbest fitness: %f\n\n", bestind, bestfit);
  fflush(f2);
}

//
//Write absolute best genome to file
//
void logBest() {
  int j;
  //best genome
  f3= fopen ("../../data/bestgenome.txt", "wt");
  fprintf(f3, "%d %d ", generation, abs_bestind);
  for(j=0; j<NB_GENES; j++)
    fprintf(f3, "%f ", pop[abs_bestind][j]);
  fprintf(f3, "\n");
  fflush(f3);
}
